#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <stdio.h>

int
reap_children(void) {
    int i = 0;

    /* Kindprozesse "beseitigen" und zhlen */
    for (;;) {
        if (wait(NULL) != -1) {
            i++;
        }
        else if (errno == ECHILD) {
            /* Es sind keine Prozesse mehr brig */
            return i;
        }
        else if (errno != EINTR) {
            /* Ein Fehler ist aufgetreten */
            return -1;
        }
    }
}

/*
 * Einen Prozess mehrfach starten.  Der Parameter `argv' enthlt
 * die gewnschte Kommandozeile, abgeschlossen mit NULL.
 */
int
run_parallel(char **argv, int anzahl_prozesse) {
    int starter[2];
    int i;
    char c;

    /* Pipe ffnen */
    if (pipe(starter) == -1) {
        /* Fehler */
        return -1;
    }
    /* Neue Prozessgruppe erzeugen */
    setpgid(0, 0);
    /* Prozesse erzeugen */
    for (i = 0; i < anzahl_prozesse; i++) {
        switch (fork()) {
            case -1: /* Fehler beim Erzeugen des Prozesses */
                fprintf(stderr, "fork failed, killing processes\n");
                /* Alle bereits erzeugten Prozesse beenden */
                signal(SIGTERM, SIG_IGN);
                kill(0, SIGTERM);
                /* Aufrumen und Programm beenden */
                reap_children();
                exit(127);
            case 0: /* Kindprozess */
                /* Eingang der Pipe schlieen */
                close(starter[1]);
                /* Am Ausgang der Pipe lesen; auf End-Of-File warten */
                while (read(starter[0], &c, 1) != 0);
                /* Programm starten */
                execvp(argv[0], argv);
                /* Fehler beim Programmstart */
                perror("execvp");
                _exit(127);
            default: /* Vaterprozess */
                break;
        }
    }
    /* Beide Enden der Pipe schlieen (Ausgang zuerst) */
    close(starter[0]);
    close(starter[1]);
    /* Kindprozesse starten an dieser Stelle */
    /* Auf Beendigung warten; Anzahl zurckgeben */
    return reap_children();
}

/*
 * Beispielprogramm
 */
int
main(int argc, char **argv) {
    char *program[] = { "echo", "Hello, world!", NULL };

    run_parallel(program, 10);
    exit(0);
}
